#include "MPU6886.h"
#include <Kalman.h>
#include "FastLED.h"
#include <WiFi.h>
#include <WebServer.h>
#include <Preferences.h>
#include "unit_rolleri2c.hpp"


UnitRollerI2C RollerI2C0, RollerI2C1, RollerI2C2;  // Create a UNIT_ROLLERI2C object

#define led_pin 27
#define NUM_LEDS 25
CRGB leds[NUM_LEDS];

MPU6886 IMU;


WebServer server(80);

const char ssid[] = "S-3 Roller";  // SSID
const char pass[] = "password";   // password

const IPAddress ip(192, 168, 122, 133);      // IPアドレス
const IPAddress subnet(255, 255, 255, 0); // サブネットマスク


unsigned long oldTime = 0, loopTime, nowTime;
float dt;

float Kp = 60;
float Kd = 65;
float Kw = 1.2;
float IDRS = 0.1;
int delayTime = 5;

float offsetX = 0.0, offsetY = 0.0;


int GetUP = 0;
float MX = 0.0, MY = 0.0, M0 = 0.0, M1 = 0.0, M2 = 0.0;
float AjX = 0.0, AjY = 0.0;
float velX = 0.0, velY = 0.0;


float accX = 0, accY = 0, accZ = 0;
float gyroX = 0, gyroY = 0, gyroZ = 0;


float theta_X = 0.0, theta_Y = 0.0;
float theta_Xdot = 0.0, theta_Ydot = 0.0, theta_Zdot = 0.0;

Kalman kalmanX, kalmanY;
float kalAngleX, kalAngleY, kalAngleDotX, kalAngleDotY;

Preferences preferences;


//加速度センサから傾きデータ取得 [deg]
void get_theta() {
  IMU.getAccelData(&accX,&accY,&accZ);
  //傾斜角導出 単位はdeg
  theta_X = atan2(accY , -accZ) * 180.0/PI  + offsetX;
  theta_Y = -atan2(-accX , -accZ) * 180.0/PI + offsetY;
}

//Y軸 角速度取得
void get_gyro_data() {
  IMU.getGyroData(&gyroX,&gyroY,&gyroZ);
  theta_Xdot = -gyroX;
  theta_Ydot = gyroY;
}


//ブラウザ表示
void handleRoot() {
  String temp = "<!DOCTYPE html> \n<html lang=\"ja\">";
  temp += "<head>";
  temp += "<meta charset=\"utf-8\">";
  temp += "<title>S-3 Roller</title>";
  temp += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">";
  temp += "<style>";
  temp += ".container{";
  temp += "  max-width: 500px;";
  temp += "  margin: auto;";
  temp += "  text-align: center;";
  temp += "  font-size: 1.2rem;";
  temp += "}";
  temp += "span,.pm{";
  temp += "  display: inline-block;";
  temp += "  border: 1px solid #ccc;";
  temp += "  width: 50px;";
  temp += "  height: 30px;";
  temp += "  vertical-align: middle;";
  temp += "  margin-bottom: 20px;";
  temp += "}";
  temp += "span{";
  temp += "  width: 120px;";
  temp += "}";
  temp += "button{";
  temp += "  width: 100px;";
  temp += "  height: 40px;";
  temp += "  font-weight: bold;";
  temp += "  margin-bottom: 20px;";
  temp += "}";
  temp += "</style>";
  temp += "</head>";

  temp += "<body>";
  temp += "<div class=\"container\">";
  temp += "<h3>S-3 Roller</h3>";

  
  //offsetX
  temp += "offsetX<br>";
  temp += "<a class=\"pm\" href=\"/offsetXm\">-</a>";
  temp += "<span>" + String(offsetX) + "</span>";
  temp += "<a class=\"pm\" href=\"/offsetXp\">+</a><br>";

  //offsetY
  temp += "offsetY<br>";
  temp += "<a class=\"pm\" href=\"/offsetYm\">-</a>";
  temp += "<span>" + String(offsetY) + "</span>";
  temp += "<a class=\"pm\" href=\"/offsetYp\">+</a><br>";

  

  //Kp
  temp += "Kp<br>";
  temp += "<a class=\"pm\" href=\"/KpM\">-</a>";
  temp += "<span>" + String(Kp) + "</span>";
  temp += "<a class=\"pm\" href=\"/KpP\">+</a><br>";

  //Kd
  temp += "Kd<br>";
  temp += "<a class=\"pm\" href=\"/KdM\">-</a>";
  temp += "<span>" + String(Kd) + "</span>";
  temp += "<a class=\"pm\" href=\"/KdP\">+</a><br>";

  //Kw
  temp += "Kw<br>";
  temp += "<a class=\"pm\" href=\"/KwM\">-</a>";
  temp += "<span>" + String(Kw) + "</span>";
  temp += "<a class=\"pm\" href=\"/KwP\">+</a><br>";

  //IDRS
  temp += "IDRS<br>";
  temp += "<a class=\"pm\" href=\"/IDRSm\">-</a>";
  temp += "<span>" + String(IDRS) + "</span>";
  temp += "<a class=\"pm\" href=\"/IDRSp\">+</a><br>";

  //delayTime
  temp += "delayTime<br>";
  temp += "<a class=\"pm\" href=\"/delayTimeM\">-</a>";
  temp += "<span>" + String(delayTime) + "</span>";
  temp += "<a class=\"pm\" href=\"/delayTimeP\">+</a><br>";


  temp += "</div>";
  temp += "</body>";
  server.send(200, "text/HTML", temp);
}


void offsetXm() {
  if (offsetX >= -10) {
    offsetX -= 0.1;
    preferences.putFloat("offsetX", offsetX);
  }
  handleRoot();
}

void offsetXp() {
  if (offsetX <= 10) {
    offsetX += 0.1;
    preferences.putFloat("offsetX", offsetX);
  }
  handleRoot();
}


void offsetYm() {
  if (offsetY >= -10) {
    offsetY -= 0.1;
    preferences.putFloat("offsetY", offsetY);
  }
  handleRoot();
}

void offsetYp() {
  if (offsetY <= 10) {
    offsetY += 0.1;
    preferences.putFloat("offsetY", offsetY);
  }
  handleRoot();
}



void KpM() {
  if (Kp >= 0.1) {
    Kp -= 1;
    preferences.putFloat("Kp", Kp);
  }
  handleRoot();
}

void KpP() {
  if (Kp <= 300) {
    Kp += 1;
    preferences.putFloat("Kp", Kp);
  }
  handleRoot();
}

void KdM() {
  if (Kd >= 0.1) {
    Kd -= 1;
    preferences.putFloat("Kd", Kd);
  }
  handleRoot();
}

void KdP() {
  if (Kd <= 300) {
    Kd += 1;
    preferences.putFloat("Kd", Kd);
  }
  handleRoot();
}

void KwM() {
  if (Kw >= 0) {
    Kw -= 0.1;
    preferences.putFloat("Kw", Kw);
  }
  handleRoot();
}

void KwP() {
  if (Kw <= 300) {
    Kw += 0.1;
    preferences.putFloat("Kw", Kw);
  }
  handleRoot();
}

void IDRSm() {
  if (IDRS >= 0.0) {
    IDRS -= 0.01;
    preferences.putFloat("IDRS", IDRS);
  }
  handleRoot();
}

void IDRSp() {
  if (IDRS <= 10) {
    IDRS += 0.01;
    preferences.putFloat("IDRS", IDRS);
  }
  handleRoot();
}


void delayTimeM() {
  if (delayTime > 0) {
    delayTime -= 1;
    preferences.putInt("delayTime", delayTime);
  }
  handleRoot();
}

void delayTimeP() {
  if (delayTime <= 30) {
    delayTime += 1;
    preferences.putInt("delayTime", delayTime);
  }
  handleRoot();
}


//Core0
void displayIMU(void *pvParameters) {
  FastLED.addLeds<WS2812B, led_pin, GRB>(leds, NUM_LEDS);
  FastLED.setBrightness(20);

  IMU.Init();

  //フルスケールレンジ
  IMU.SetAccelFsr(IMU.AFS_2G);
  IMU.SetGyroFsr(IMU.GFS_250DPS);

  get_theta();
  kalmanX.setAngle(theta_X);
  kalmanY.setAngle(theta_Y);


  delay(100); 


  WiFi.softAP(ssid, pass);           // SSIDとパスの設定
  delay(100);                        // 追記：このdelayを入れないと失敗する場合がある
  WiFi.softAPConfig(ip, ip, subnet); // IPアドレス、ゲートウェイ、サブネットマスクの設定

  IPAddress myIP = WiFi.softAPIP();  // WiFi.softAPIP()でWiFi起動

  server.on("/", handleRoot);
  
  server.on("/offsetXm", offsetXm);
  server.on("/offsetXp", offsetXp);
  server.on("/offsetYm", offsetYm);
  server.on("/offsetYp", offsetYp);
  

  server.on("/KpP", KpP);
  server.on("/KpM", KpM);
  server.on("/KdP", KdP);
  server.on("/KdM", KdM);
  server.on("/KwP", KwP);
  server.on("/KwM", KwM);
  server.on("/IDRSp", IDRSp);
  server.on("/IDRSm", IDRSm);
  server.on("/delayTimeP", delayTimeP);
  server.on("/delayTimeM", delayTimeM);

  server.begin();

  for (;;) {
    server.handleClient();

    nowTime = micros();
    loopTime = nowTime - oldTime;
    oldTime = nowTime;
    
    dt = (float)loopTime / 1000000.0; //sec

    get_theta();
    get_gyro_data();
  
      
    //カルマンフィルタ 姿勢 傾き
    kalAngleX = kalmanX.getAngle(theta_X, theta_Xdot, dt);
    kalAngleY = kalmanY.getAngle(theta_Y, theta_Ydot, dt);
    
    //カルマンフィルタ 姿勢 角速度
    kalAngleDotX = kalmanX.getRate();
    kalAngleDotY = kalmanY.getRate();

    Serial.print(kalAngleX);
    Serial.print("\t");
    Serial.println(kalAngleY);


    //LED表示
    for(int i = 0; i < NUM_LEDS; i++){
      leds[i] = CRGB::Black;
    }


    if(kalAngleX > 20.0){
      for(int i = 0; i < 5; i++){
        leds[i] = CRGB::Red;
      }
    }else if(kalAngleX <= 20.0 && kalAngleX > 12.0){
      for(int i = 0; i < 5; i++){
        leds[i] = CRGB::Green;
      }
    }else if(kalAngleX <= 12.0 && kalAngleX > 4.0){
      for(int i = 0; i < 5; i++){
        leds[i + 5] = CRGB::Green;
      }
    }else if(kalAngleX <= 4.0 && kalAngleX > 1.0){
      for(int i = 0; i < 5; i++){
        leds[i + 10] = CRGB::Green;
      }
    }else if(abs(kalAngleX) <= 1.0){
      for(int i = 0; i < 5; i++){
        leds[i + 10] = CRGB::Blue;
      }
    }else if(kalAngleX >= -4.0 && kalAngleX < -1.0){
      for(int i = 0; i < 5; i++){
        leds[i + 10] = CRGB::Green;
      }
    }else if(kalAngleX >= -12.0 && kalAngleX < -4.0){
      for(int i = 0; i < 5; i++){
        leds[i + 15] = CRGB::Green;
      }
    }else if(kalAngleX >= -20.0 && kalAngleX < -12.0){
      for(int i = 0; i < 5; i++){
        leds[i + 20] = CRGB::Green;
      }
    }else if(kalAngleX < -20.0){
      for(int i = 0; i < 5; i++){
        leds[i + 20] = CRGB::Red;
      }
    }
    
      
    if(kalAngleY > 20.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5] = CRGB::Red;
      }
    }else if(kalAngleY <= 20.0 && kalAngleY > 12.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5] = CRGB::Green;
      }
    }else if(kalAngleY <= 12.0 && kalAngleY > 4.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 1] = CRGB::Green;
      }
    }else if(kalAngleY <= 4.0 && kalAngleY > 1.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 2] = CRGB::Green;
      }
    }else if(abs(kalAngleY) <= 1.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 2] = CRGB::Blue;
      }
    }else if(kalAngleY >= -4.0 && kalAngleY < -1.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 2] = CRGB::Green;
      }
    }else if(kalAngleY >= -12.0 && kalAngleY < -4.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 3] = CRGB::Green;
      }
    }else if(kalAngleY >= -20.0 && kalAngleY < -12.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 4] = CRGB::Green;
      }
    }else if(kalAngleY < -20.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 4] = CRGB::Red;
      }
    }
    
    FastLED.show();
        
    delay(delayTime);
    
    disableCore0WDT();
  }
}



void setup() {
  Serial.begin(115200);

  RollerI2C0.begin(&Wire, 0x64, 26, 32, 400000); 
  RollerI2C1.begin(&Wire, 0x65, 26, 32, 400000); 
  RollerI2C2.begin(&Wire, 0x66, 26, 32, 400000); 


  //パラメータ初期値取得
  preferences.begin("parameter", false);

  offsetX = preferences.getFloat("offsetX", offsetX);
  offsetY = preferences.getFloat("offsetY", offsetY);
  
  Kp = preferences.getFloat("Kp", Kp);
  Kd = preferences.getFloat("Kd", Kd);
  Kw = preferences.getFloat("Kw", Kw);
  IDRS = preferences.getFloat("IDRS", IDRS);
  delayTime = preferences.getInt("delayTime", delayTime);


  //回転位置検出 タスク
  xTaskCreatePinnedToCore(
    displayIMU
    ,  "display"   // A name just for humans
    ,  4096  // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  1  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL
    ,  0);


  delay(1000);

  RollerI2C0.setMode(ROLLER_MODE_CURRENT); 
  RollerI2C1.setMode(ROLLER_MODE_CURRENT); 
  RollerI2C2.setMode(ROLLER_MODE_CURRENT); 
  RollerI2C0.setOutput(1);
  RollerI2C1.setOutput(1);
  RollerI2C2.setOutput(1);
}


void loop() {
  if (fabs(kalAngleX) < 1.0 && fabs(kalAngleY) < 1.0 && GetUP == 0) {
    GetUP = 80;
  }
  

  if(GetUP == 80){
    //ブレーキ
    if(fabs(kalAngleX) > 20.0 || fabs(kalAngleY) > 20.0){
      RollerI2C0.setCurrent(0);
      RollerI2C1.setCurrent(0);
      RollerI2C2.setCurrent(0);
      AjX = 0.0;
      AjY = 0.0;
      GetUP = 0;
    }
        
    //モータ回転
    velY =  -RollerI2C0.getSpeedReadback() / 100.0;
    AjY +=  IDRS * velY / 10000.0;
    MY = (Kp * (kalAngleY + AjY) + Kd * kalAngleDotY / 10.0 + Kw * velY / 10.0) * 1000.0;

    velX =  0.5 / 100.0 * (-RollerI2C1.getSpeedReadback() + RollerI2C2.getSpeedReadback());
    AjX +=  IDRS * velX / 10000.0;
    MX = (Kp * (kalAngleX + AjX) + Kd * kalAngleDotX / 10.0 + Kw * velX / 10.0) * 1000.0;

    M0 = MY;
    M1 = -0.5 * MY + 0.866 * MX;
    M2 = 0.5 * MY + 0.866 * MX;
    
    RollerI2C0.setCurrent(-M0);
    RollerI2C1.setCurrent(-M1);
    RollerI2C2.setCurrent(M2);
    
  }else if(GetUP == 0){
    RollerI2C0.setCurrent(0);
    RollerI2C1.setCurrent(0);
    RollerI2C2.setCurrent(0);
  }
}
